﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace HeuristicLab.Algorithms.Bandits {
  public class UCBNormalPolicy : BanditPolicy {
    private readonly int[] tries;
    private readonly double[] sumReward;
    private readonly double[] sumSqrReward;
    private int totalTries = 0;
    public UCBNormalPolicy(int numActions)
      : base(numActions) {
      this.tries = new int[numActions];
      this.sumReward = new double[numActions];
      this.sumSqrReward = new double[numActions];
    }

    public override int SelectAction() {
      Debug.Assert(Actions.Any());
      int bestAction = -1;
      double bestQ = double.NegativeInfinity;
      foreach (var a in Actions) {
        if (totalTries == 0 || tries[a] == 0 || tries[a] < Math.Ceiling(8 * Math.Log(totalTries))) return a;
        var avgReward = sumReward[a] / tries[a];
        var q = avgReward
          + Math.Sqrt(16 * ((sumSqrReward[a] - tries[a] * Math.Pow(avgReward, 2)) / (tries[a] - 1)) * (Math.Log(totalTries - 1) / tries[a]));
        if (q > bestQ) {
          bestQ = q;
          bestAction = a;
        }
      }
      return bestAction;
    }
    public override void UpdateReward(int action, double reward) {
      Debug.Assert(Actions.Contains(action));
      totalTries++;
      tries[action]++;
      sumReward[action] += reward;
      sumSqrReward[action] += reward * reward;
    }

    public override void DisableAction(int action) {
      base.DisableAction(action);
      totalTries -= tries[action];
      tries[action] = -1;
      sumReward[action] = 0;
      sumSqrReward[action] = 0;
    }

    public override void Reset() {
      base.Reset();
      totalTries = 0;
      Array.Clear(tries, 0, tries.Length);
      Array.Clear(sumReward, 0, sumReward.Length);
      Array.Clear(sumSqrReward, 0, sumSqrReward.Length);
    }
  }
}
